// Probabilistic model of the Human-Robot handover task.
//
// Used for the paper: 
// An Assurance-based Approach to Verification and Validation of Human--Robot Teams
// Webster, Western, Araiza-Illan, Dixon, Eder, Fisher and Pipe
//  
// Includes support for verification of Requirement 1 - Property 6.
//
// Matt Webster, University of Liverpool, matt@liverpool.ac.uk. Aug 2016.
// Works with PRISM 4.2.1.

dtmc


////////////////////////////////////////////////////////////////
// HUMAN MODEL
////////////////////////////////////////////////////////////////

// Human states
const int start 			= 0;
const int valuesAreSet			= 1;
const int activatedRobot		= 2;
const int responding			= 3;
const int setGPL			= 4;
const int holdingObject			= 5;
const int fixingObject			= 6;
const int gazeSet			= 7;
const int pressureSet			= 8;
const int locationSet			= 9;
const int onTask			= 10;
const int tired				= 11;
const int bored				= 12;
const int offTask			= 13;


const int timeMax			= 1000;
const int ortMax			= 200;



// measures time in 0.1 second intervals
module timekeeper
	
	time: [0..timeMax] init 0;  // time in tenths of a second
	objectReleaseTimer: [0..ortMax] init 0;
	// use synchronisations to update the clock
	[activateRobot] !GPLWasOk & time<=(timeMax-40) -> (time'=time+40);
	[informHumanOfHandoverStart] !GPLWasOk & time<=(timeMax-60) -> (time'=time+60);
	[movingHand] !GPLWasOk & time<=(timeMax-100) -> (time'=time+100);
	[graspingObject] !GPLWasOk & time<=(timeMax-50) -> (time'=time+50);
	[humanIsReady] !GPLWasOk & time<=(timeMax-10) -> (time'=time+10);
	[senseGaze] !GPLWasOk & time<=(timeMax-1) -> (time'=time+1);
	[sensePressure] !GPLWasOk & time<=(timeMax-1) -> (time'=time+1);
	[senseLocation] !GPLWasOk & time<=(timeMax-1) -> (time'=time+1);
	[tick] !GPLWasOk & time<=(timeMax-1) -> (time'=time+1);
	[GPLOkSet] !GPLWasOk & time<=(timeMax-1)-> (time'=time+1) & (objectReleaseTimer'=0);

	[activateRobot] GPLWasOk  & time<=(timeMax-40) & objectReleaseTimer<=(ortMax-40) -> (time'=time+40) & (objectReleaseTimer'=objectReleaseTimer+40);
	[informHumanOfHandoverStart] GPLWasOk  & time<=(timeMax-60) & objectReleaseTimer<=(ortMax-60) -> (time'=time+60) & (objectReleaseTimer'=objectReleaseTimer+60);
	[movingHand] GPLWasOk  & time<=(timeMax-100) & objectReleaseTimer<=(ortMax-100) -> (time'=time+100) & (objectReleaseTimer'=objectReleaseTimer+100);
	[graspingObject] GPLWasOk  & time<=(timeMax-50) & objectReleaseTimer<=(ortMax-50)-> (time'=time+50) & (objectReleaseTimer'=objectReleaseTimer+50);
	[humanIsReady] GPLWasOk  & time<=(timeMax-10) & objectReleaseTimer<=(ortMax-10) -> (time'=time+10) & (objectReleaseTimer'=objectReleaseTimer+10);
	[senseGaze] GPLWasOk  & time<=(timeMax-1) & objectReleaseTimer<=(ortMax-1) -> (time'=time+1) & (objectReleaseTimer'=objectReleaseTimer+1);
	[sensePressure] GPLWasOk  & time<=(timeMax-1) & objectReleaseTimer<=(ortMax-1) -> (time'=time+1) & (objectReleaseTimer'=objectReleaseTimer+1);
	[senseLocation] GPLWasOk & time<=(timeMax-1) & objectReleaseTimer<=(ortMax-1) -> (time'=time+1) & (objectReleaseTimer'=objectReleaseTimer+1);
	[tick] GPLWasOk & time<=(timeMax) & objectReleaseTimer<=(ortMax) -> (time'=time+1) & (objectReleaseTimer'=objectReleaseTimer+1);
	[openGripper] GPLWasOk & time<=(timeMax-20) & objectReleaseTimer<=(ortMax-20) -> (time'=time+20) & (objectReleaseTimer'=objectReleaseTimer+20);
endmodule


// Human probabilities
const double pDisengages		= 0;
//const double pTired			= 0.01;
const double pStaysOnTask		= 1-pDisengages;

// Human module
// Clocks modelled: clk_isTired, clk_HRI
module human

	humanState : [0..99] init start;
	[activateRobot] humanState=start -> (humanState'=activatedRobot);
	[tick] humanState=activatedRobot -> (humanState'=activatedRobot);
	[informHumanOfHandoverStart] humanState=activatedRobot -> (humanState'=responding);
	[humanIsReady] humanState=responding -> (humanState'=setGPL);
	[] humanState=setGPL -> (humanState'=setGPL);
	[tick] humanState=setGPL -> pDisengages: (humanState'=offTask) + pStaysOnTask: (humanState'=setGPL);


endmodule




////////////////////////////////////////////////////////////////
// GAZE MODEL
////////////////////////////////////////////////////////////////

// Gaze states
const int gazeOk			= 100;
const int gazeNotOk			= 101;
// Gaze probabilities
const double pGazeOk			= 1;

// Gaze module
module gaze

	gazeState : [0..199] init start;
	// human either looks at the table leg, or doesn't. The value of gazeState can be set multiple times.
	[] true -> pGazeOk: (gazeState'=gazeOk) + (1-pGazeOk): (gazeState'=gazeNotOk);
	

endmodule


////////////////////////////////////////////////////////////////
// PRESSURE MODEL
////////////////////////////////////////////////////////////////


// Pressure states
const int pressureOk			= 110;
const int pressureNotOk			= 111;
// Pressure probabilities
const double pPressureOk		= 1;
// Pressure module
module pressure

	pressureState : [0..199] init start;
	// human is either taking the weight of the table leg, or it isn't
	[] true -> pPressureOk: (pressureState'=pressureOk) + (1-pPressureOk): (pressureState'=pressureNotOk);
	
endmodule


////////////////////////////////////////////////////////////////
// LOCATION MODEL
////////////////////////////////////////////////////////////////


// Location states
const int locationOk			= 112;
const int locationNotOk			= 113;
// Location probabilities
const double pLocationOk		= 1;
// Location module
module location

	locationState : [0..199] init start;
	// human's hand is either in the right location, or it isn't.
	[] true -> pLocationOk: (locationState'=locationOk) + (1-pLocationOk): (locationState'=locationNotOk);


endmodule


////////////////////////////////////////////////////////////////
// SENSORS MODEL
////////////////////////////////////////////////////////////////

// Sensor states
const int null 				= 1000;

// Sensor probabilities of true positives (TP), false negatives (FN), true negatives (TN) and false positives (FP) for each sensor.
const double pGazeFN			= 0.00;			// based on experiment
const double pGazeTP			= 1-pGazeFN;   		
const double pGazeFP			= 0.00;   		// based on experiment
const double pGazeTN			= 1-pGazeFP;
const double pPressureFN		= 0.071428571;   		// based on experiment
const double pPressureTP		= 1-pPressureFN;
const double pPressureFP		= 0.00;   	// based on experiment
const double pPressureTN		= 1-pPressureFP;
const double pLocationFN		= 0.030612245;   		// based on experiment
const double pLocationTP		= 1-pLocationFN;
const double pLocationFP		= 0.00;		// based on experiment
const double pLocationTN		= 1-pLocationFP;   	
	


// Sensor module
module gazeSensor
	gazeSensorState : [0..1000] init null;
	gazeSensorSet: bool init false;
	
	// assign values to the gaze, position and location sensors. TP, FN, TN and FP are all possible for each sensor, according to the probabilities set above.
	[senseGaze] robotState=waitForGPLUpdate & gazeState=gazeOk & !gazeSensorSet -> pGazeFN: (gazeSensorState'=gazeNotOk) & (gazeSensorSet'=true) + pGazeTP: (gazeSensorState'=gazeOk) & (gazeSensorSet'=true);
	[senseGaze] robotState=waitForGPLUpdate & gazeState=gazeNotOk & !gazeSensorSet -> pGazeTN: (gazeSensorState'=gazeNotOk) & (gazeSensorSet'=true) + pGazeFP: (gazeSensorState'=gazeOk) & (gazeSensorSet'=true);
endmodule


module pressureSensor
	pressureSensorState : [0..1000] init null;
	pressureSensorSet: bool init false;
	
	// assign values to the gaze, position and location sensors. TP, FN, TN and FP are all possible for each sensor, according to the probabilities set above.
	// These values can be set multiple times as long as the robot is waiting for sensor updates (robotState=waitForGPLUpdate)
	[sensePressure] robotState=waitForGPLUpdate & pressureState=pressureOk & !pressureSensorSet -> pPressureFN: (pressureSensorState'=pressureNotOk) & (pressureSensorSet'=true) 
													+ pPressureTP: (pressureSensorState'=pressureOk) & (pressureSensorSet'=true);
	[sensePressure] robotState=waitForGPLUpdate & pressureState=pressureNotOk & !pressureSensorSet -> pPressureTN: (pressureSensorState'=pressureNotOk) & (pressureSensorSet'=true) 
													+ pPressureFP: (pressureSensorState'=pressureOk) & (pressureSensorSet'=true);
endmodule


module locationSensor
	locationSensorState : [0..1000] init null;
	locationSensorSet: bool init false;
	
	// assign values to the gaze, position and location sensors. TP, FN, TN and FP are all possible for each sensor, according to the probabilities set above.
	[senseLocation] robotState=waitForGPLUpdate & locationState=locationOk & !locationSensorSet -> pLocationFN: (locationSensorState'=locationNotOk) & (locationSensorSet'=true) 
													+ pLocationTP: (locationSensorState'=locationOk) & (locationSensorSet'=true);
	[senseLocation] robotState=waitForGPLUpdate & locationState=locationNotOk & !locationSensorSet -> pLocationTN: (locationSensorState'=locationNotOk) & (locationSensorSet'=true) 
													+ pLocationFP: (locationSensorState'=locationOk) & (locationSensorSet'=true);	
endmodule

////////////////////////////////////////////////////////////////
// ROBOT MODEL
////////////////////////////////////////////////////////////////

// Robot holding states
const int nothing			= 900;
const int leg				= 901;
const int gripperFailure		= 902;


// Robot states
const int moveHandToObjectLocation 	= 1100;
const int graspObject			= 1101;
const int moveHandToHandoverLocation 	= 1102;
const int informedHumanOfHandoverStart	= 1103;
const int waitForGazeUpdate		= 1104;
const int waitForPressureUpdate		= 1105;
const int waitForLocationUpdate 	= 1106;
const int waitForGPLUpdate 		= 1107;
const int decideReleaseObject 		= 1108;
// Equivalent states which have the same value ////
// This is because we want to rename releasedObject to GPLOk but we don't want to break any properties including releasedObject
const int releasedObject		= 1109;
const int GPLOk				= 1109;
///////////////////////////////////////////////////
const int RholdingObject		= 1110;
const int waiting			= 1111;
const int interactionDone		= 1112;
const int GPLBad			= 1113;
const int handoverSuccessful		= 1114;
//const int InteractionDone		= 1115;
const int handoverUnsuccessful		= 1116;
const int motionError			= 1117;
const int timedOut			= 1118;





// Robot module
module robot
	robotState: [1100..1199] init waiting;
	handContents : [0..1000] init nothing;
	testedTimeout: bool init false;
	GPLWasOk: bool init false;

	[activateRobot] robotState=waiting -> (robotState'=moveHandToObjectLocation);


	[movingHand] robotState=moveHandToObjectLocation -> (robotState'=graspObject) & (handContents'=leg);
	[graspingObject] robotState=graspObject -> (robotState'=moveHandToHandoverLocation);
	[informHumanOfHandoverStart] robotState=moveHandToHandoverLocation -> (robotState'=informedHumanOfHandoverStart);
	[humanIsReady] robotState=informedHumanOfHandoverStart -> (robotState'=waitForGPLUpdate);
	
	[GPLOkSet] robotState=waitForGPLUpdate & humanState=setGPL & gazeSensorState=gazeOk & pressureSensorState=pressureOk & locationSensorState=locationOk -> (robotState'=GPLOk) & (GPLWasOk'=true);

	[tick] robotState=waitForGPLUpdate & humanState=setGPL & (gazeSensorState=gazeNotOk | pressureSensorState=pressureNotOk | locationSensorState=locationNotOk)  
		-> (robotState'=waitForGPLUpdate);
	


	[tick] robotState=waitForGPLUpdate & (humanState=tired | humanState=bored | humanState=offTask) -> (robotState'=interactionDone);

	[openGripper] robotState=GPLOk -> (handContents'=nothing) & (robotState'=handoverSuccessful);


	[] time=timeMax -> (robotState'=timedOut);



endmodule





// END OF FILE
